local helpers = require "spec.helpers"
local jwt = require "resty.jwt"
local cjson = require "cjson.safe"

local PLUGIN_NAME = "universal-jwt"

local function has_value(tab, val)
  for index, value in ipairs(tab) do
    if value == val then
      return true
    end
  end
  return false
end

for _, strategy in helpers.each_strategy() do
  describe("universal-jwt plugin (access)", function()
    local proxy_client
    local admin_client
    local consumer_username = "bob"
    local consumer_groups = {"admin", "owner"}
    local bp, db, dao = helpers.get_db_utils(strategy, nil, {PLUGIN_NAME})

    setup(function()
      local service1 = assert(bp.services:insert { 
          name = "service-1", 
      })

      local api1 = assert(bp.routes:insert { 
          name = "api-1", 
          hosts = { "test1.com" },
          service = { id = service1.id }
      })

      local api2 = assert(bp.routes:insert { 
          name = "api-2", 
          hosts = { "test2.com" }
          -- upstream_url = helpers.mock_upstream_url
      })

      assert(bp.plugins:insert {
        name   = "key-auth",
        route = {
          id = api1.id
        },
      })

      local consumer1 = assert(bp.consumers:insert {
        username = consumer_username,
      })
      assert(bp.keyauth_credentials:insert {
        key         = "apikey1",
        consumer = { id = consumer1.id }
      })

      assert(bp.plugins:insert {
        route = {
          id = api1.id
        },
        name = PLUGIN_NAME,
      })

      assert(bp.acls:insert {
        group = consumer_groups[1],
        consumer = { id = consumer1.id }
      })

      assert(bp.acls:insert {
        group = consumer_groups[2],
        consumer = { id = consumer1.id }
      })

      assert(helpers.start_kong({
        database = "postgres",
        nginx_conf = "spec/fixtures/custom_nginx.template",
        plugins = "bundled,universal-jwt"
      }))

      admin_client = helpers.admin_client()
      proxy_client = helpers.proxy_client()
    end)

    teardown(function()
      if admin_client then
        admin_client:close()
      end

      if proxy_client then
        proxy_client:close()
      end

      helpers.stop_kong()
    end)

    before_each(function()
      proxy_client = helpers.proxy_client()
    end)

    after_each(function()
      if proxy_client then
        proxy_client:close()
      end
    end)

    describe("request", function()
      it("with key auth plugin adds the 'Authorization' header", function()
        local r = assert(proxy_client:send {
          method = "GET",
          path = "/request?apikey=apikey1",
          headers = {
            host = "test1.com",
            ["x-localz-deviceid"] = "device1"
          }
        })
        assert.response(r).has.status(200)
        local auth_header = assert.request(r).has.header("Authorization")
        local token = auth_header:sub(8)
        local header, claims, signature = token:match("([^.]*).([^.]*).(.*)")
        claims = cjson.decode(ngx.decode_base64(claims))
        assert.are.equals(consumer_username, claims.iss)
        assert(has_value(claims.scopes.roles, consumer_groups[1]))
        assert(has_value(claims.scopes.roles, consumer_groups[2]))
      end)
      it("without key auth plugin does nothing", function()
        local r = assert(proxy_client:send {
          method = "GET",
          path = "/request",
          headers = {
            host = "test2.com"
          }
        })
        assert.response(r).has.status(200)
        assert.request(r).has_not.header("Authorization")
      end)
    end)
  end)
end
